#include <stdio.h>
#include <signal.h>
#include <string.h>

#define DEVDIR "/dev/"
#define TRUE  (0==0)
#define FALSE (0!=0)
#define ERR_OK		0
#define ERR_LOCKED      1
#define ERR_USER        2
#define ERR_EXTNAL      5

int usage(void);
void cleanup(int);
char *getpidcmdline(int);

char *progname;
int quit;
char lockfile[256];

main(
int argc,
char *argv[])
{
    char *cp;

    progname = (cp=strrchr(argv[0],'/')) ? cp+1 : argv[0];

    if (argc != 3)
        return(usage());

    if (strcmp("on", argv[2]) == 0) 
        return(lock(argv[1], 1));
    else if (strcmp("off", argv[2]) == 0)
        return(lock(argv[1], 0));
    else 
        return(usage());
}

int lock(
char *device,
int onoff)
{
    int i;
    int slashcnt;

    if (strncmp(DEVDIR, device, strlen(DEVDIR))) {
        fprintf(stderr, "%s: ERROR: invalid device: %s\n", progname, device);
        return(ERR_USER);
    }
    for (i=0,slashcnt=0; i<strlen(device); i++)
        if (device[i] == '/')
            slashcnt++;
    if (slashcnt != 2) {
        fprintf(stderr, "%s: ERROR: invalid device: %s\n", progname, device);
        return(ERR_USER);
    }

    sprintf(lockfile, "/usr/spool/uucp/LCK..%s", device+strlen(DEVDIR));
    return(onoff ? lockon(lockfile) : lockoff(lockfile));
}

int lockon(
char *lockfile)
{
    int pid;
    FILE *fp;
    char *cmdline;

    if (readpid(lockfile, &pid) != 0)
        return(forklock(lockfile));
    
    if ((cmdline=getpidcmdline(pid)) != NULL) {
        fprintf(stderr, "%s: ERROR: locked by %s\n", progname, cmdline);
        return(ERR_LOCKED);

    } else if (unlink(lockfile)) {
        fprintf(stderr, "%s: ERROR: can't unlink stale lock: %s\n", progname, lockfile);
        return(ERR_EXTNAL);

    } else {
        fprintf(stderr, "%s: WARNING: stale lock removed: %s\n", progname, lockfile);
        return(forklock(lockfile));
    }
}
      
int lockoff(
char *lockfile)
{
    int pid;
    FILE *fp;
    char *cmdline;

    if (readpid(lockfile, &pid) != 0) {
        fprintf(stderr, "%s: WARNING: not locked\n", progname);
        return(ERR_OK);

    } else if ((cmdline=getpidcmdline(pid)) == NULL && unlink(lockfile)) {
        fprintf(stderr, "%s: ERROR: can't unlink stale lock: %s\n", progname, lockfile);
        return(ERR_EXTNAL);

    } else if (cmdline == NULL) {
        fprintf(stderr, "%s: WARNING: stale lock removed: %s\n", progname, lockfile);
        return(ERR_OK);
    }

    if (strcmp(cmdline, progname) == 0) {
        kill(pid, SIGTERM);
        return(ERR_OK);

    } else {
        fprintf(stderr, "%s: ERROR: locked by %s\n", progname, cmdline);
        return(ERR_LOCKED);
    }
}

int forklock()
{
    FILE *fp;
    int pid;

    if ((pid=fork()) < 0) {
        fprintf(stderr, "%s: ERROR: can't fork\n", progname);
        return(ERR_EXTNAL);

    } else if (pid != 0)
        return(writepid(lockfile, pid));

    signal(SIGINT, cleanup);
    signal(SIGTERM, cleanup);
    for (quit=FALSE; !quit;)
        sleep(3600);
    return(ERR_OK);
}

int readpid(
char *lockfile,
int *pidp)
{ 
    FILE *fp;

    if ((fp=fopen(lockfile, "r")) == NULL)
        return(ERR_EXTNAL);
    fread(pidp, sizeof(int), 1, fp);
    fclose(fp);
    return(ERR_OK);
}

int writepid(
char *lockfile,
int pid)
{ 
    FILE *fp;

    if ((fp=fopen(lockfile, "w")) == NULL) {
        fprintf(stderr, "%s: ERROR: can't write lock file: %s\n", progname, lockfile);
        return(ERR_EXTNAL);
    }
    fwrite(&pid, sizeof(int), 1, fp);
    fclose(fp);
    return(ERR_OK);
}

int usage()
{
    fprintf(stderr, "Usage: %s device [ on | off ]\n", progname);
    return(ERR_USER);
}

void cleanup(
int sig)
{
    unlink(lockfile);
    quit = TRUE;
}

char *getpidcmdline(
int pid)
{
    static char cmdline[256];
    char linbuf[256];
    FILE *fp;
    char *cp;

    sprintf(linbuf, "/proc/%d/cmdline", pid);
    if ((fp=fopen(linbuf, "r")) == NULL)
        return(NULL);
    fgets(cmdline, sizeof(cmdline), fp);
    fclose(fp);
    return((cp=strrchr(cmdline, '/')) ? cp+1 : cmdline);
}
